"""
Things involved in this firewall:
1. Rate limiting
2. IP Throttling
3. Blocklist (After the rate max limit is exceeded)
4. Timeout
5. Connection limit (One connection per IP)
6. DOS attack prevention (Block IP after exceeding rate limit for a certain number of times)
7. Logging
8. Saving blocklist to file
"""

import socket
from subprocess import Popen, STDOUT, PIPE
from threading import Thread, Lock
import time
import re
import json
import logging

# Setup logging
logging.basicConfig(filename='firewall.log', level=logging.INFO, format='%(asctime)s - %(message)s')

BLOCKLIST = []
WINDOW_TIME = 60
RATE_LIMIT = 10
REQUEST_TIMEOUT = 30
PROCESSING_TIME = 5
active_connection = {}
lock = Lock()
valid_math_expression = re.compile(r'^[0-9+\-*/().\s\^]+$')

BLOCKLIST_FILE = 'blocklist.json'

# Load blocklist from file
def load_blocklist():
    global BLOCKLIST
    try:
        with open(BLOCKLIST_FILE, 'r') as f:
            BLOCKLIST = json.load(f)
    except FileNotFoundError:
        BLOCKLIST = []

# Save blocklist to file
def save_blocklist():
    with open(BLOCKLIST_FILE, 'w') as f:
        json.dump(BLOCKLIST, f)

# Initialize blocklist
load_blocklist()

MAX_CONNECTIONS = 100
IP_THROTTLE_LIMIT = 5
IP_THROTTLE_TIME = 10
DOS_THRESHOLD = 3  # Number of times rate limit can be exceeded before blocking

connection_count = 0
ip_throttle = {}
dos_attempts = {}

def add_to_blocklist(ip_address):
    global BLOCKLIST
    if ip_address not in BLOCKLIST:
        BLOCKLIST.append(ip_address)
        save_blocklist()
        logging.info(f"Added {ip_address} to blocklist.")
        print(f"Added {ip_address} to blocklist.")

def start_new_math_thread(conn, addr):
    global connection_count
    ip_address = addr[0]
    with lock:
        if ip_address in BLOCKLIST:
            print(f"Connection was closed because the {ip_address} is blocklisted.")
            conn.close()
            return
        if ip_address in active_connection:
            print(f"Connection blocked from {ip_address} (already connected).")
            conn.sendall(b"Only one connection per IP is allowed.")
            conn.close()
            return
        if connection_count >= MAX_CONNECTIONS:
            print("Server is too busy. Connection refused.")
            conn.sendall(b"Server is too busy. Please try again later.")
            conn.close()
            return
        if ip_address in ip_throttle and time.time() - ip_throttle[ip_address] < IP_THROTTLE_TIME:
            print(f"Connection throttled from {ip_address}.")
            conn.sendall(b"Too many requests. Please try again later(15sec)")
            conn.close()
            return
        active_connection[ip_address] = conn
        connection_count += 1
    
    try:
        t = MathServerCommunicationThread(conn, addr)
        t.start()
    except Exception as e:
        print(f"Error at starting thread for the connection {ip_address}: {e}")
        with lock:
            del active_connection[ip_address]
            connection_count -= 1

class ProcessOutputThread(Thread):
    def __init__(self, proc, conn):
        Thread.__init__(self)
        self.proc = proc
        self.conn = conn

    def run(self):
        while True:
            try:
                output = self.proc.stdout.readline()
                if output:
                    self.conn.sendall(output.encode()+b"$")

                else:
                    break
            except Exception as e:
                print("Output thread error:", e)
                break

class MathServerCommunicationThread(Thread):
    def __init__(self, conn, addr):
        Thread.__init__(self)
        self.conn = conn
        self.addr = addr
        self.request_time = []
        self.last_request_time = time.time()
        self.conn.settimeout(REQUEST_TIMEOUT)

    def run(self):
        global connection_count, dos_attempts
        try:
            print(f"{self.addr[0]} connected with port {self.addr[1]}")
            self.conn.sendall(b"Simple Math Server developed by LAHTP\n\nEnter any math expressions, and I will answer you. Type 'quit' or 'exit' to close.\n\n$")

            p = Popen(['bc'], stdout=PIPE, stdin=PIPE, stderr=STDOUT, text=True)
            output_thread = ProcessOutputThread(p, self.conn)
            output_thread.start()

            while True:
                try:
                    current_time = time.time()
                    if current_time - self.last_request_time > REQUEST_TIMEOUT:  # timeout
                        self.conn.sendall(b"Connection terminated due to inactivity.")
                        print(f"Connection from {self.addr[0]} timed out.")
                        break

                    # Check rate limiting
                    self.request_time = [t for t in self.request_time if current_time - t < WINDOW_TIME]
                    if len(self.request_time) >= RATE_LIMIT:
                        self.conn.sendall(b"Exceeding rate limit, please wait before sending another request.")
                        logging.info(f"Rate limit exceeded for {self.addr[0]}.")
                        print(f"Rate limit exceeded for {self.addr[0]}.")
                        dos_attempts[self.addr[0]] = dos_attempts.get(self.addr[0], 0) + 1
                        if dos_attempts[self.addr[0]] >= DOS_THRESHOLD:
                            add_to_blocklist(self.addr[0])
                        break  # Break the loop immediately after sending the rate limit message

                    data = self.conn.recv(1024)
                    if not data:
                        break
                    self.last_request_time = time.time()  # Update last request time
                    query = data.decode().strip()
                    if query.lower() in ['quit', 'exit']:
                        self.conn.sendall(b"Closing connection. Bye!")
                        break
                    if not valid_math_expression.match(query):
                        self.conn.sendall(b"Invalid expression. Please enter a valid math expression.\n")
                        continue

                    self.request_time.append(current_time)
                    query += '\n'

                    p.stdin.write(query)
                    p.stdin.flush()

                except socket.timeout:
                    self.conn.sendall(b"Connection timed out due to inactivity.\n")
                    break
                except Exception as e:
                    error_message = f"Error with connection {self.addr[0]}: {e}"
                    print(error_message)
                    self.conn.sendall(f"{error_message}\n".encode())
                    break
        finally:
            with lock:
                if self.addr[0] in active_connection:
                    del active_connection[self.addr[0]]
                connection_count -= 1
                ip_throttle[self.addr[0]] = time.time()
            self.conn.close()
            print(f"Connection terminated from {self.addr[0]}.")

# Save blocklist on exit
import atexit
atexit.register(save_blocklist)

HOST = '0.0.0.0'
PORT = 3074

s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
s.bind((HOST, PORT))
s.listen()

while True:
    conn, addr = s.accept()
    start_new_math_thread(conn, addr)

s.close()
